@@ -1,7 +1,11 @@
 # Author:  Lisandro Dalcin
 # Contact: dalcinl@gmail.com
 """Management of MPI worker processes."""
+# pylint: disable=broad-except
+# pylint: disable=too-many-lines
+# pylint: disable=protected-access
 # pylint: disable=missing-docstring
+# pylint: disable=import-outside-toplevel
 
 import os
 import sys
@@ -14,6 +18,7 @@ import threading
 import collections
 
 from .. import MPI
+from ._core import BrokenExecutor
 
 
 # ---
@@ -26,7 +31,7 @@ def serialized(function):
         with serialized.lock:
             return function(*args, **kwargs)
     return wrapper
-serialized.lock = None
+serialized.lock = None  # type: ignore[attr-defined]
 
 
 def setup_mpi_threads():
@@ -35,29 +40,43 @@ def setup_mpi_threads():
         if thread_level is None:
             thread_level = MPI.Query_thread()
             setup_mpi_threads.thread_level = thread_level
-            if thread_level < MPI.THREAD_MULTIPLE:  # pragma: no cover
+            if thread_level < MPI.THREAD_MULTIPLE:
                 serialized.lock = threading.Lock()
-    if thread_level < MPI.THREAD_SERIALIZED:  # pragma: no cover
+    if thread_level < MPI.THREAD_SERIALIZED:
         warnings.warn("The level of thread support in MPI "
                       "should be at least MPI_THREAD_SERIALIZED",
                       RuntimeWarning, 2)
-setup_mpi_threads.lock = threading.Lock()
-setup_mpi_threads.thread_level = None
+setup_mpi_threads.lock = threading.Lock()  # type: ignore[attr-defined]
+setup_mpi_threads.thread_level = None      # type: ignore[attr-defined]
 
 
 # ---
 
 
-if sys.version_info[0] == 2:     # pragma: no cover
-    def sys_exception():
-        exc = sys.exc_info()[1]
-        return exc
-else:                            # pragma: no cover
+if sys.version_info[0] >= 3:
+
     def sys_exception():
         exc = sys.exc_info()[1]
         exc.__traceback__ = None
         return exc
 
+else:  # pragma: no cover
+
+    def sys_exception():
+        exc = sys.exc_info()[1]
+        return exc
+
+
+def os_environ_get(name, default=None):
+    varname = 'MPI4PY_FUTURES_{}'.format(name)
+    if varname not in os.environ:
+        oldname = 'MPI4PY_{}'.format(name)
+        if oldname in os.environ:  # pragma: no cover
+            message = "Environment variable {} is deprecated, use {}"
+            warnings.warn(message.format(oldname, varname), DeprecationWarning)
+            return os.environ[oldname]
+    return os.environ.get(varname, default)
+
 
 # ---
 
@@ -65,7 +84,7 @@ else:                            # pragma: no cover
 BACKOFF = 0.001
 
 
-class Backoff(object):
+class Backoff:
 
     def __init__(self, seconds=BACKOFF):
         self.tval = 0.0
@@ -91,7 +110,7 @@ class Stack(collections.deque):
     pop = collections.deque.pop
 
 
-THREADS_QUEUES = weakref.WeakKeyDictionary()
+THREADS_QUEUES = weakref.WeakKeyDictionary()  # type: weakref.WeakKeyDictionary
 
 
 def join_threads(threads_queues=THREADS_QUEUES):
@@ -102,10 +121,13 @@ def join_threads(threads_queues=THREADS_QUEUES):
         thread.join()
 
 
-atexit.register(join_threads)
+try:
+    threading._register_atexit(join_threads)  # type: ignore[attr-defined]
+except AttributeError:  # pragma: no cover
+    atexit.register(join_threads)
 
 
-class Pool(object):
+class Pool:
 
     def __init__(self, executor, manager, *args):
         self.size = None
@@ -113,21 +135,23 @@ class Pool(object):
         self.queue = queue = Queue()
         self.exref = weakref.ref(executor, lambda _, q=queue: q.put(None))
 
-        args = (self,) + args
-        kwargs = executor._options  # pylint: disable=protected-access
-        thread = threading.Thread(target=manager, args=args, kwargs=kwargs)
-        thread.daemon = True
+        args = (self, executor._options) + args
+        thread = threading.Thread(target=manager, args=args)
         self.thread = thread
 
         setup_mpi_threads()
+        try:
+            threading._register_atexit
+        except AttributeError:  # pragma: no cover
+            thread.daemon = True
         thread.start()
         THREADS_QUEUES[thread] = queue
 
     def wait(self):
         self.event.wait()
 
-    def push(self, task):
-        self.queue.put(task)
+    def push(self, item):
+        self.queue.put(item)
 
     def done(self):
         self.queue.put(None)
@@ -135,19 +159,78 @@ class Pool(object):
     def join(self):
         self.thread.join()
 
+    def setup(self, size):
+        self.size = size
+        self.event.set()
+        return self.queue
 
-def setup_pool(pool, num_workers):
-    pool.size = num_workers
-    pool.event.set()
-    return pool.queue
+    def cancel(self, handler=None):
+        queue = self.queue
+        while True:
+            try:
+                item = queue.pop()
+            except LookupError:
+                break
+            if item is None:
+                queue.put(None)
+                break
+            future, _ = item
+            if handler:
+                handler(future)
+            else:
+                future.cancel()
+            del item, future
+
+    def broken(self, message):
+        lock = None
+        executor = self.exref()
+        if executor is not None:
+            executor._broken = message
+            if not executor._shutdown:
+                lock = executor._lock
+
+        def handler(future):
+            if future.set_running_or_notify_cancel():
+                exception = BrokenExecutor(message)
+                future.set_exception(exception)
+
+        if lock:
+            lock.acquire()
+        try:
+            self.cancel(handler)
+        finally:
+            if lock:
+                lock.release()
 
 
-def _manager_thread(pool, **options):
+def initialize(options):
+    initializer = options.pop('initializer', None)
+    initargs = options.pop('initargs', ())
+    initkwargs = options.pop('initkwargs', {})
+    if initializer is not None:
+        try:
+            initializer(*initargs, **initkwargs)
+            return True
+        except BaseException:
+            return False
+    return True
+
+
+def _manager_thread(pool, options):
     size = options.pop('max_workers', 1)
-    queue = setup_pool(pool, size)
+    queue = pool.setup(size)
+
+    def init():
+        if not initialize(options):
+            pool.broken("initializer failed")
+            return False
+        return True
 
     def worker():
         backoff = Backoff(options.get('backoff', BACKOFF))
+        if not init():
+            queue.put(None)
+            return
         while True:
             try:
                 item = queue.pop()
@@ -179,31 +262,39 @@ def _manager_thread(pool, **options):
     queue.pop()
 
 
-def _manager_comm(pool, comm, **options):
+def _manager_comm(pool, options, comm, full=True):
+    assert comm != MPI.COMM_NULL
+    serialized(client_sync)(comm, options, full)
     size = comm.Get_remote_size()
+    queue = pool.setup(size)
     workers = Stack(reversed(range(size)))
-    queue = setup_pool(pool, size)
-    client(comm, 0, workers, queue, **options)
+    if not client_init(comm, options):
+        pool.broken("initializer failed")
+        serialized(client_close)(comm)
+        return
+    client_exec(comm, options, 0, workers, queue)
+    serialized(client_close)(comm)
 
 
-def _manager_spawn(pool, **options):
+def _manager_split(pool, options, comm, root):
+    comm = serialized(comm_split)(comm, root)
+    _manager_comm(pool, options, comm, full=False)
+
+
+def _manager_spawn(pool, options):
     pyexe = options.pop('python_exe', None)
     pyargs = options.pop('python_args', None)
     nprocs = options.pop('max_workers', None)
     info = options.pop('mpi_info', None)
     comm = serialized(client_spawn)(pyexe, pyargs, nprocs, info)
-    serialized(client_sync)(comm, options)
-    _manager_comm(pool, comm, **options)
-    serialized(client_close)(comm)
+    _manager_comm(pool, options, comm)
 
 
-def _manager_service(pool, **options):
+def _manager_service(pool, options):
     service = options.pop('service', None)
     info = options.pop('mpi_info', None)
     comm = serialized(client_connect)(service, info)
-    serialized(client_sync)(comm, options)
-    _manager_comm(pool, comm, **options)
-    serialized(client_close)(comm)
+    _manager_comm(pool, options, comm)
 
 
 def ThreadPool(executor):
@@ -211,6 +302,11 @@ def ThreadPool(executor):
     return Pool(executor, _manager_thread)
 
 
+def SplitPool(executor, comm, root):
+    # pylint: disable=invalid-name
+    return Pool(executor, _manager_split, comm, root)
+
+
 def SpawnPool(executor):
     # pylint: disable=invalid-name
     return Pool(executor, _manager_spawn)
@@ -223,7 +319,6 @@ def ServicePool(executor):
 
 def WorkerPool(executor):
     # pylint: disable=invalid-name
-    # pylint: disable=protected-access
     if SharedPool is not None:
         return SharedPool(executor)
     if 'service' in executor._options:
@@ -232,62 +327,6 @@ def WorkerPool(executor):
         return SpawnPool(executor)
 
 
-# ---
-
-
-def get_comm_world():
-    return MPI.COMM_WORLD
-
-
-def comm_split(comm, root=0):
-    assert not comm.Is_inter()
-    assert comm.Get_size() > 1
-    assert 0 <= root < comm.Get_size()
-    rank = comm.Get_rank()
-
-    if MPI.Get_version() >= (2, 2):
-        allgroup = comm.Get_group()
-        if rank == root:
-            group = allgroup.Incl([root])
-        else:
-            group = allgroup.Excl([root])
-        allgroup.Free()
-        intracomm = comm.Create(group)
-        group.Free()
-    else:  # pragma: no cover
-        color = 0 if rank == root else 1
-        intracomm = comm.Split(color, key=0)
-
-    if rank == root:
-        local_leader = 0
-        remote_leader = 0 if root else 1
-    else:
-        local_leader = 0
-        remote_leader = root
-    intercomm = intracomm.Create_intercomm(
-        local_leader, comm, remote_leader, tag=0)
-    intracomm.Free()
-
-    return intercomm
-
-
-def _manager_split(pool, comm, root, **options):
-    comm = serialized(comm_split)(comm, root)
-    _manager_comm(pool, comm, **options)
-    serialized(client_close)(comm)
-
-
-def SplitPool(executor, comm, root):
-    # pylint: disable=invalid-name
-    return Pool(executor, _manager_split, comm, root)
-
-
-def server_main_split(comm, root, **options):
-    comm = comm_split(comm, root)
-    server(comm, **options)
-    server_close(comm)
-
-
 # ---
 
 SharedPool = None  # pylint: disable=invalid-name
@@ -300,16 +339,24 @@ def _set_shared_pool(obj):
     SharedPool = obj
 
 
-def _manager_shared(pool, comm, tag, workers, **options):
+def _manager_shared(pool, options, comm, tag, workers):
     # pylint: disable=too-many-arguments
     if tag == 0:
         serialized(client_sync)(comm, options)
     size = comm.Get_remote_size()
-    queue = setup_pool(pool, size)
-    client(comm, tag, workers, queue, **options)
+    queue = pool.setup(size)
+    if tag == 0:
+        if not client_init(comm, options):
+            pool.broken("initializer failed")
+            return
+    if tag >= 1:
+        if options.get('initializer') is not None:
+            pool.broken("cannot run initializer")
+            return
+    client_exec(comm, options, tag, workers, queue)
 
 
-class SharedPoolCtx(object):
+class SharedPoolCtx:
     # pylint: disable=too-few-public-methods
 
     def __init__(self):
@@ -350,11 +397,14 @@ class SharedPoolCtx(object):
         if self.comm != MPI.COMM_NULL:
             if self.on_root:
                 if next(self.counter) == 0:
-                    client_sync(self.comm, dict(main=False))
+                    options = dict(main=False)
+                    client_sync(self.comm, options)
+                    client_init(self.comm, options)
                 client_close(self.comm)
             else:
                 options = server_sync(self.comm)
-                server(self.comm, **options)
+                server_init(self.comm)
+                server_exec(self.comm, options)
                 server_close(self.comm)
         if not self.on_root:
             join_threads(self.threads)
@@ -390,11 +440,9 @@ def barrier(comm):
         MPI.Request.Waitall(sendreqs)
 
 
-def client_sync(comm, options):
+def bcast_send(comm, data):
     assert comm.Is_inter()
     assert comm.Get_size() == 1
-    barrier(comm)
-    data = _sync_get_data(options)
     if MPI.VERSION >= 2:
         comm.bcast(data, MPI.ROOT)
     else:  # pragma: no cover
@@ -405,8 +453,41 @@ def client_sync(comm, options):
             for pid in range(size)])
 
 
-def client(comm, tag, worker_pool, task_queue, **options):
+def bcast_recv(comm):
+    assert comm.Is_inter()
+    assert comm.Get_remote_size() == 1
+    if MPI.VERSION >= 2:
+        data = comm.bcast(None, 0)
+    else:  # pragma: no cover
+        tag = MPI.COMM_WORLD.Get_attr(MPI.TAG_UB)
+        data = comm.recv(None, 0, tag)
+    return data
+
+
+# ---
+
+
+def client_sync(comm, options, full=True):
+    assert comm.Is_inter()
+    assert comm.Get_size() == 1
+    barrier(comm)
+    if full:
+        options = _sync_get_data(options)
+    bcast_send(comm, options)
+
+
+def client_init(comm, options):
+    serialized(bcast_send)(comm, _init_get_data(options))
+    sbuf = bytearray([False])
+    rbuf = bytearray([False])
+    serialized(comm.Allreduce)(sbuf, rbuf, op=MPI.LAND)
+    success = bool(rbuf[0])
+    return success
+
+
+def client_exec(comm, options, tag, worker_pool, task_queue):
     # pylint: disable=too-many-locals
+    # pylint: disable=too-many-statements
     assert comm.Is_inter()
     assert comm.Get_size() == 1
     assert tag >= 0
@@ -457,12 +538,12 @@ def client(comm, tag, worker_pool, task_queue, **options):
             pid = worker_pool.pop()
         except LookupError:  # pragma: no cover
             task_queue.add(item)
-            return
+            return False
 
         future, task = item
         if not future.set_running_or_notify_cancel():
             worker_pool.put(pid)
-            return
+            return False
 
         try:
             request = comm_isend(task, pid, tag)
@@ -470,6 +551,7 @@ def client(comm, tag, worker_pool, task_queue, **options):
         except BaseException:
             worker_pool.put(pid)
             future.set_exception(sys_exception())
+        return None
 
     while True:
         if task_queue and worker_pool:
@@ -497,20 +579,27 @@ def client_close(comm):
         comm.Free()
 
 
-def server_sync(comm):
+def server_sync(comm, full=True):
     assert comm.Is_inter()
     assert comm.Get_remote_size() == 1
     barrier(comm)
-    if MPI.VERSION >= 2:
-        data = comm.bcast(None, 0)
-    else:  # pragma: no cover
-        tag = MPI.COMM_WORLD.Get_attr(MPI.TAG_UB)
-        data = comm.recv(None, 0, tag)
-    options = _sync_set_data(data)
+    options = bcast_recv(comm)
+    if full:
+        options = _sync_set_data(options)
     return options
 
 
-def server(comm, **options):
+def server_init(comm):
+    options = bcast_recv(comm)
+    success = initialize(options)
+    sbuf = bytearray([success])
+    rbuf = bytearray([True])
+    comm.Allreduce(sbuf, rbuf, op=MPI.LAND)
+    assert bool(rbuf[0]) is False
+    return success
+
+
+def server_exec(comm, options):
     assert comm.Is_inter()
     assert comm.Get_remote_size() == 1
 
@@ -573,8 +662,48 @@ def server_close(comm):
 # ---
 
 
+def get_comm_world():
+    return MPI.COMM_WORLD
+
+
+def comm_split(comm, root=0):
+    assert not comm.Is_inter()
+    assert comm.Get_size() > 1
+    assert 0 <= root < comm.Get_size()
+    rank = comm.Get_rank()
+
+    if MPI.Get_version() >= (2, 2):
+        allgroup = comm.Get_group()
+        if rank == root:
+            group = allgroup.Incl([root])
+        else:
+            group = allgroup.Excl([root])
+        allgroup.Free()
+        intracomm = comm.Create(group)
+        group.Free()
+    else:  # pragma: no cover
+        color = 0 if rank == root else 1
+        intracomm = comm.Split(color, key=0)
+
+    if rank == root:
+        local_leader = 0
+        remote_leader = 0 if root else 1
+    else:
+        local_leader = 0
+        remote_leader = root
+    intercomm = intracomm.Create_intercomm(
+        local_leader, comm, remote_leader, tag=0)
+    intracomm.Free()
+    return intercomm
+
+
+# ---
+
+
+MAIN_RUN_NAME = '__worker__'
+
+
 def import_main(mod_name, mod_path, init_globals, run_name):
-    # pylint: disable=protected-access
     import types
     import runpy
 
@@ -587,7 +716,7 @@ def import_main(mod_name, mod_path, init_globals, run_name):
         # pylint: disable=too-few-public-methods
         def __init__(self, mod_name):
             # pylint: disable=no-member
-            super(TempModulePatch, self).__init__(mod_name)
+            super().__init__(mod_name)
             assert self.module.__name__ == run_name
             self.module = module
 
@@ -600,6 +729,8 @@ def import_main(mod_name, mod_path, init_globals, run_name):
         if mod_name:  # pragma: no cover
             runpy.run_module(mod_name, run_name=run_name, alter_sys=True)
         elif mod_path:  # pragma: no branch
+            if not getattr(sys.flags, 'isolated', 0):  # pragma: no branch
+                sys.path[0] = os.path.realpath(os.path.dirname(mod_path))
             runpy.run_path(mod_path, run_name=run_name)
         sys.modules['__main__'] = sys.modules[run_name] = module
     except:  # pragma: no cover
@@ -610,18 +741,23 @@ def import_main(mod_name, mod_path, init_globals, run_name):
         runpy._TempModule = TempModule
 
 
-MAIN_RUN_NAME = '__worker__'
-
-
 def _sync_get_data(options):
     main = sys.modules['__main__']
     sys.modules.setdefault(MAIN_RUN_NAME, main)
     import_main_module = options.pop('main', True)
 
     data = options.copy()
+    data.pop('initializer', None)
+    data.pop('initargs', None)
+    data.pop('initkwargs', None)
+
     if import_main_module:
-        spec = getattr(main, '__spec__', None)
-        name = getattr(spec, 'name', None)
+        if sys.version_info[0] >= 3:
+            spec = getattr(main, '__spec__', None)
+            name = getattr(spec, 'name', None)
+        else:  # pragma: no cover
+            loader = getattr(main, '__loader__', None)
+            name = getattr(loader, 'fullname', None)
         path = getattr(main, '__file__', None)
         if name is not None:  # pragma: no cover
             data['@main:mod_name'] = name
@@ -647,10 +783,17 @@ def _sync_set_data(data):
     return data
 
 
+def _init_get_data(options):
+    keys = ('initializer', 'initargs', 'initkwargs')
+    vals = (None, (), {})
+    data = dict((k, options.pop(k, v)) for k, v in zip(keys, vals))
+    return data
+
+
 # ---
 
 
-def check_recursive_spawn():  # pragma: no cover
+def _check_recursive_spawn():  # pragma: no cover
     if not hasattr(import_main, 'sentinel'):
         return
     main_name, main_path = import_main.sentinel
@@ -691,6 +834,9 @@ FLAG_OPT_MAP = {
     'verbose': 'v',
     'quiet': 'q',
     'bytes_warning': 'b',
+    # 'dev_mode': 'Xdev',
+    # 'utf8_mode': 'Xutf8',
+    # 'warn_default_encoding': 'Xwarn_default_encoding',
     # Python 2
     'division_warning': 'Qwarn',
     'division_new': 'Qnew',
@@ -718,8 +864,9 @@ def get_python_flags():
 
 
 def get_max_workers():
-    if 'MPI4PY_MAX_WORKERS' in os.environ:
-        return int(os.environ['MPI4PY_MAX_WORKERS'])
+    max_workers = os_environ_get('MAX_WORKERS')
+    if max_workers is not None:
+        return int(max_workers)
     if MPI.UNIVERSE_SIZE != MPI.KEYVAL_INVALID:  # pragma: no branch
         universe_size = MPI.COMM_WORLD.Get_attr(MPI.UNIVERSE_SIZE)
         if universe_size is not None:  # pragma: no cover
@@ -728,11 +875,15 @@ def get_max_workers():
     return 1
 
 
+def get_spawn_module():
+    return __spec__.parent + '.server'
+
+
 def client_spawn(python_exe=None,
                  python_args=None,
                  max_workers=None,
                  mpi_info=None):
-    check_recursive_spawn()
+    _check_recursive_spawn()
     if python_exe is None:
         python_exe = sys.executable
     if python_args is None:
@@ -743,7 +894,7 @@ def client_spawn(python_exe=None,
         mpi_info = dict(soft='1:{}'.format(max_workers))
 
     args = get_python_flags() + list(python_args)
-    args.extend(['-m', __package__ + '.server'])
+    args.extend(['-m', get_spawn_module()])
     info = MPI.Info.Create()
     info.update(mpi_info)
     comm = MPI.COMM_SELF.Spawn(python_exe, args, max_workers, info)
@@ -754,26 +905,26 @@ def client_spawn(python_exe=None,
 # ---
 
 
-SERVICE = __package__
+SERVICE = __spec__.parent
 SERVER_HOST = 'localhost'
 SERVER_BIND = ''
 SERVER_PORT = 31415
 
 
 def get_service():
-    return os.environ.get('MPI4PY_SERVICE', SERVICE)
+    return os_environ_get('SERVICE', SERVICE)
 
 
 def get_server_host():
-    return os.environ.get('MPI4PY_SERVER_HOST', SERVER_HOST)
+    return os_environ_get('SERVER_HOST', SERVER_HOST)
 
 
 def get_server_bind():
-    return os.environ.get('MPI4PY_SERVER_BIND', SERVER_BIND)
+    return os_environ_get('SERVER_BIND', SERVER_BIND)
 
 
 def get_server_port():
-    return int(os.environ.get('MPI4PY_SERVER_PORT', SERVER_PORT))
+    return int(os_environ_get('SERVER_PORT', SERVER_PORT))
 
 
 def client_lookup(address):
@@ -869,31 +1020,37 @@ def server_accept(service, mpi_info=None,
 # ---
 
 
-def server_main_comm(comm):
+def server_main_comm(comm, full=True):
     assert comm != MPI.COMM_NULL
-    options = server_sync(comm)
-    server(comm, **options)
+    options = server_sync(comm, full)
+    server_init(comm)
+    server_exec(comm, options)
     server_close(comm)
 
 
+def server_main_split(comm, root):
+    comm = comm_split(comm, root)
+    server_main_comm(comm, full=False)
+
+
 def server_main_spawn():
     comm = MPI.Comm.Get_parent()
     server_main_comm(comm)
 
 
-def server_main_accept():
+def server_main_service():
     from getopt import getopt
     longopts = ['bind=', 'port=', 'service=', 'info=']
     optlist, _ = getopt(sys.argv[1:], '', longopts)
-    options = {opt[2:] : val for opt, val in optlist}
+    optdict = {opt[2:]: val for opt, val in optlist}
 
-    if 'bind' in options or 'port' in options:
-        bind = options.get('bind') or get_server_bind()
-        port = options.get('port') or get_server_port()
+    if 'bind' in optdict or 'port' in optdict:
+        bind = optdict.get('bind') or get_server_bind()
+        port = optdict.get('port') or get_server_port()
         service = (bind, int(port))
     else:
-        service = options.get('service') or get_service()
-    info = options.get('info', '').split(',')
+        service = optdict.get('service') or get_service()
+    info = optdict.get('info', '').split(',')
     info = dict(k_v.split('=', 1) for k_v in info if k_v)
 
     comm = server_accept(service, info)
@@ -907,7 +1064,7 @@ def server_main():
         if comm != MPI.COMM_NULL:
             server_main_spawn()
         else:
-            server_main_accept()
+            server_main_service()
     except:
         set_abort_status(1)
         raise
